//
//  DiceGL.m
//  DiceView
//
//  Created by kunii on 10/03/29.
//  Copyright 2010 Apple Inc. All rights reserved.
//

#import "DiceGL.h"

@implementation DiceGL

/*
	UIImageをテクスチャにマッピングする。
*/
-(void)map:(UIImage*)inImage texture:(GLuint)inTexture {
	CGImageRef cgImage = [inImage CGImage];
	int width = CGImageGetWidth(cgImage);
	int height = CGImageGetHeight(cgImage);
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	void* data = malloc(width * height * 4);
	CGContextRef cgContext = CGBitmapContextCreate(data, width, height, 8, width * 4, colorSpace, kCGImageAlphaNoneSkipLast | kCGBitmapByteOrder32Big);
	if (cgContext != NULL) {
		CGContextDrawImage(cgContext, CGRectMake(0.0f, 0.0f, width, height), cgImage);		
		glBindTexture(GL_TEXTURE_2D, inTexture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
		GLenum err = glGetError();
		if (err != GL_NO_ERROR)
			NSLog(@"Error uploading texture. glError: 0x%04X", err);
		CGContextRelease(cgContext);
	}	
	free(data);
	CGColorSpaceRelease(colorSpace);
}

/*
	初期画像（サイコロ）にする。
*/
-(void)initImage {
	for (int i = 0; i < 6; i++) {
		[self map:defaultImages[i] texture:texture[i]];
	}
}

/*
	初期化
	
	初期画像（サイコロ）、テクスチャの用意。
*/
-(id)init {
	if ([super init]) {
		for (int i = 0; i < 6; i++) {
			defaultImages[i] = [UIImage imageWithContentsOfFile:[[NSBundle mainBundle] pathForResource:[NSString stringWithFormat:@"%d", i + 1] ofType:@"png"]];
			[defaultImages[i] retain];
			if (defaultImages[i] == nil) {
				NSLog(@"Failed to load %d.png", i);
				return 0;
			}	
		}
		glGenTextures(6, texture);
		[self initImage];
		photos = [[NSMutableArray alloc] init];
		angle = 0;
		curtImageIndex[0] = 0;
		curtImageIndex[1] = 1;
		movingPhase = movingPhase_Idle;		
		xAxisRotateAngle = 0;
		xAxisRotateDeltaAngle = 0.2;
	}
	return self;
}

- (void)dealloc {
	for (int i = 0; i < 6; i++) {
		[defaultImages[i] release];
	}
	glDeleteTextures(6, texture);
	[photos release];
	[super dealloc];
}

//	平面を構成する頂点
static const GLfloat vertices[][3] = {
	{-1.0, 1.0, 1.0}, 
	{1.0, 1.0, 1.0}, 
	{-1.0, -1.0, 1.0}, 
	{1.0, -1.0, 1.0}
};

//	法線ベクトル
static const GLfloat normals[][3] = {				
	{0.0, 0.0, 1.0},
	{0.0, 0.0, 1.0},
	{0.0, 0.0, 1.0},
	{0.0, 0.0, 1.0}
};

//	平面へのテクスチャのマップ座標
static const GLfloat squareTexCoords[][2] = {
	{0.0, 0.0},
	{1.0, 0.0},
	{0.0, 1.0},
	{1.0, 1.0}
};

//	6面分の回転指定
static const GLfloat rotates[][4] = {
	{0, 0.0, 1.0, 0.0},
	{90, 0.0, 1.0, 0.0},
	{180, 0.0, 1.0, 0.0},
	{-90, 0.0, 1.0, 0.0},
	{90, 1.0, 0.0, 0.0},
	{-90, 1.0, 0.0, 0.0}
};

/*
	平面を描画する。
*/
-(void)drawPlane:(int)part {
	glPushMatrix(); 
	if (part > 0)
		glRotatef(rotates[part][0], rotates[part][1], rotates[part][2], rotates[part][3]); 

	glEnable(GL_TEXTURE_2D);

	glTexCoordPointer(2, GL_FLOAT, 0, squareTexCoords);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);	//	GL_CLAMP_TO_EDGEだと繰り返さない。
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	if (maxIndex == 0) {
		//	サイコロ描画
		glBindTexture(GL_TEXTURE_2D, texture[part]);
	} else {
		glBindTexture(GL_TEXTURE_2D, texture[(part % 2)]);
	}
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glDisable(GL_TEXTURE_2D);
	glPopMatrix(); 
}


//	サイコロを振る。ただし振って動いている最中なら止める。
#if THROW_ANIMATION
-(void)prepareRotate:(double)inDelta {
	marginIndex = rand() % 3 + 1;					//	じらし用余白。
	int turnangle = rand() % 90;					//	オーナーの写真をずらす角度
	if (maxIndex > 0) {
		//	どの画像をオーナーにするかを決める。
		int curtIndex = ((int)angle / 90) % maxIndex;	//	現在の番号。最後に止まった位置から始めるため。
		int resultIndex = rand() % maxIndex;		//	これがマスターになる。
		if (turnangle >= 45) {						//	45度以上の場合四捨五入されるので、その前に一つ減らしておく。
			resultIndex--;
			if (resultIndex == -1) {				//	ただし-1になる場合は、最大値にする。
				resultIndex = maxIndex - 1;
			}
		}
		resultIndex += (maxIndex - curtIndex);
		int appendIndex = marginIndex + (1 * maxIndex * fabs(inDelta));
		int rotateIndex = (appendIndex + maxIndex - 1) / maxIndex * maxIndex;	//	かならずmaxIndexの倍数にする。
		stopAngle = (resultIndex + rotateIndex - marginIndex) * 90;
	} else { 
		//	画像がないので、適当な番号で止めればいい。
		stopAngle = 2 * 6 * fabs(inDelta) * 90;
	}
	//	3秒間で2週 + ランダム回転するように初速を計算する。
	double time = 3 * 60;							//	3秒間分の呼び出し回数。drawは1/60秒ごとに呼ばれるので（EAGLView.m参照）。
	deltaAngle = stopAngle / time;					//	初速設定。（2週 + ランダム） / 3秒間分の呼び出し回数　
	deltaDeltaAngle = -(deltaAngle * deltaAngle) / (2.0 * stopAngle);	//	加速度
	stopAngle += (marginIndex * 90) + turnangle;
	stopAngle += angle;
	rotationDirection = (inDelta >= 0.0)?1.0:-1.0;	//	画面上の回転方向を決める
	deltaTime = 0;									//	回転経過時間を初期化
}

/*
	ダイスが振られた。
		回転中なら止める。
	3秒間で2週程度させる。
*/
-(void)throwDice:(double)inDelta {
	if (movingPhase == movingPhase_Rotate) {
		//	回転中なら現在の角度を設定する。
		stopAngle = angle;
		return;
	}
	if (movingPhase == movingPhase_Turn) {
		//	回転が行き過ぎて戻っている最中なら、最終決定角度を設定。
		angle = adjustedFinalAngle;
		return;
	}
	//	movingPhase_IdleもしくはmovingPhase_Stopなので、回転に入る。
	movingPhase = movingPhase_Rotate;
	[self prepareRotate:inDelta];
}


/*
	完了処理
		音を鳴らしたり、移譲インスタンスに完了を通知。
*/
-(void)finish {
	angle = adjustedFinalAngle;
	movingPhase = movingPhase_Stop;
/*
	int index = (adjustedFinalAngle / 90);
	if (maxIndex > 0)
		index = index % maxIndex;
	else
		index = -1;
	[delegate determ:index];
*/
}

//	目標の角度に向かって回転速度を落としながら回転。
-(void)rotation {
	if (angle >= stopAngle) {
		//	目標の角度に到着、行き過ぎた分を戻るための計算。
		adjustedFinalAngle = ((int)stopAngle + 45) / 90 * 90;
		deltaAngle = (adjustedFinalAngle - stopAngle) / (0.5 * 60);
		if (deltaAngle == 0) {
			//	ピッタリの場所で止まってるので、これで終了。
			[self finish];
		}
		//	戻りの回転。
		movingPhase = movingPhase_Turn;
		return;
	}
	//	現在の角度に追加するべき角度を計算する。
	double append = deltaAngle + deltaDeltaAngle * deltaTime;
	if (append < 1)			//	1度以下にはしない。
		append = 1;
	if ((stopAngle - angle) < (marginIndex * 90))
		append = 1;			//	marginIndexからは1固定
	if ((stopAngle - angle) < (marginIndex / 2 * 90))
		append = 0.5;		//	marginIndex/2からは1/2固定

	angle += append;		//	新しい角度を設定
	deltaTime++;			//	時間を経過させる
}

//	行き過ぎた分を戻す逆回転処理。
-(void)turn {
	angle += deltaAngle;	//	新しい角度を設定
	if (deltaAngle > 0) {
		//	正の回転方向に戻っている。
		if (angle >= adjustedFinalAngle) {	//	そのため目標のadjustedFinalAngleより大きくなった時点で完了とみなす。
			//	戻りきったので終了。
			[self finish];
		}
	} else {
		//	負の回転方向に戻っている。
		if (angle <= adjustedFinalAngle) {	//	そのため目標のadjustedFinalAngleより小さくなった時点で完了とみなす。
			//	戻りきったので終了。
			[self finish];
		}
	}
}

#else
-(void)throwDice:(double)inDelta {
	if (movingPhase == movingPhase_Rotate) {
		movingPhase = movingPhase_Idle;
		return;
	}
	movingPhase = movingPhase_Rotate;
}

//	回転
-(void)rotation {
	angle += 10;
}
#endif



/*
	描画
*/
-(void)draw {
	//	画面クリア
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepthf(1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);

	glPushMatrix(); 
	if (movingPhase == movingPhase_Idle) {
		//	上下にゆらす
		xAxisRotateAngle += xAxisRotateDeltaAngle;
		if (xAxisRotateAngle > 15) {
			xAxisRotateDeltaAngle = -xAxisRotateDeltaAngle;
			xAxisRotateAngle = 15;
		}
		if (xAxisRotateAngle < -15) {
			xAxisRotateDeltaAngle = -xAxisRotateDeltaAngle;
			xAxisRotateAngle = -15;
		}
		glRotatef(xAxisRotateAngle, 1, 0, 0);
		glRotatef(45, 0, 1, 0);
	} else {
		//	回転
#if THROW_ANIMATION
		glRotatef(rotationDirection * angle, 0, 1, 0);
#else
		glRotatef(angle, 0, 1, 0);
#endif
	}

	if (maxIndex > 0) {
		//	左側テクスチャのチェック
		int image_index = (int)angle / 90;				//	左側画像番号
		int tx_index = image_index % 2;					//	左側textureのインディックス（image_indexの最大サイズ調整前に計算）
#ifdef CORRECT
		image_index = image_index % maxIndex;			//	ただし最大サイズを超えない。		
		if (curtImageIndex[tx_index] != image_index) {	//	画像が違うなら更新
			[self map:[photos objectAtIndex:image_index] texture:texture[tx_index]];
			curtImageIndex[tx_index] = image_index;
		}
		//	右側テクスチャのチェック
		tx_index = 1 ^ tx_index;						//	右側textureのインディックス
		image_index++;									//	変更する画像番号
		image_index = image_index % maxIndex;			//	ただし最大サイズを超えない。
		if (curtImageIndex[tx_index] != image_index) {	//	画像が違うなら更新
			[self map:[photos objectAtIndex:image_index] texture:texture[tx_index]];
			curtImageIndex[tx_index] = image_index;
		}
#else
		tx_index = 1 ^ tx_index;						//	右側textureのインディックス
		image_index++;									//	右側画像番号
		image_index = image_index % maxIndex;			//	ただし最大サイズを超えない。
		if (curtImageIndex[tx_index] != image_index) {	//	画像が違うなら更新
			[self map:[photos objectAtIndex:image_index] texture:texture[tx_index]];
			curtImageIndex[tx_index] = image_index;
		}
#endif
	}

	if (movingPhase == movingPhase_Rotate) 
		[self rotation];
#if THROW_ANIMATION
	if (movingPhase == movingPhase_Turn) 
		[self turn];
#endif

	#define RGBAVALUE 1.0f
	GLfloat material_amb[4]={RGBAVALUE, RGBAVALUE, RGBAVALUE, 1.0f};	//	素材側環境光特性	RGBA
	GLfloat material_dif[4]={RGBAVALUE, RGBAVALUE, RGBAVALUE, 1.0f};	//	素材側拡散光特性	RGBA
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_amb);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_dif);
	//	頂点指定
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, vertices);
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, 0, normals);
	//	六面体描画
	for (int i = 0; i < 6; i++) 
		[self drawPlane:i];

	glPopMatrix(); 
}

//	写真設定
-(void)setImages:(UIImage**)images max:(int)max {
	movingPhase = movingPhase_Idle;
	angle = 45;
	curtImageIndex[0] = 0;
	curtImageIndex[1] = 1;
	[photos removeAllObjects];
	[self initImage];
	maxIndex = max;
	for (int i = 0; i < maxIndex; i++) {
		[photos addObject:images[i]];
		if (i < 4) {
			[self map:images[i] texture:texture[i]];
		}
	}
	if (maxIndex == 1) {
		[self map:[photos objectAtIndex:0] texture:texture[1]];
	}
}

@end



